/* Copyright (C) 2003-2004 Jamey Sharp.
 * This file is licensed under the MIT license. See the file COPYING. */

#if HAVE_FEATURES_H
#define _GNU_SOURCE /* for PTHREAD_RECURSIVE_MUTEX_INITIALIZER_NP */
#include <features.h>
#endif

#include "Xlibint.h"
#include "locking.h"
#include "xclint.h"
#include <X11/XCB/xcbext.h>
#include <X11/XCB/xcbint.h>

#include <pthread.h>

static void _XLockDisplay(Display *dpy)
{
    pthread_mutex_lock(&XCBConnectionOfDisplay(dpy)->iolock);
    _XGetXCBBufferIf(dpy, _XBufferUnlocked);
    ++dpy->xcl->lock_count;
}

void XLockDisplay(Display* dpy)
{
    LockDisplay(dpy);
    /* We want the threads in the reply queue to all get out before
     * XLockDisplay returns, in case they have any side effects the
     * caller of XLockDisplay was trying to protect against.
     * XLockDisplay puts itself at the head of the event waiters queue
     * to wait for all the replies to come in.
     * TODO: Restore this behavior on XCB.
     */
}

static void _XUnlockDisplay(Display *dpy)
{
    --dpy->xcl->lock_count;
    _XPutXCBBufferIf(dpy, _XBufferUnlocked);
    pthread_mutex_unlock(&XCBConnectionOfDisplay(dpy)->iolock);
}

void XUnlockDisplay(Display* dpy)
{
    UnlockDisplay(dpy);
}

/* returns 0 if initialized ok, -1 if unable to allocate
   a mutex or other memory */
int _XInitDisplayLock(Display *dpy)
{
#ifdef PTHREAD_RECURSIVE_MUTEX_INITIALIZER_NP
    pthread_mutex_t lock = PTHREAD_RECURSIVE_MUTEX_INITIALIZER_NP;
    XCBConnectionOfDisplay(dpy)->iolock = lock;
#else
    pthread_mutexattr_t attr;
    pthread_mutexattr_init(&attr);
    pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_RECURSIVE);
    pthread_mutex_init(&XCBConnectionOfDisplay(dpy)->iolock, &attr);
    pthread_mutexattr_destroy(&attr);
#endif

    dpy->lock_fns = (struct _XLockPtrs*)Xmalloc(sizeof(struct _XLockPtrs));
    if (dpy->lock_fns == NULL)
	return -1;

    dpy->lock = 0;
    dpy->lock_fns->lock_display = _XLockDisplay;
    dpy->lock_fns->unlock_display = _XUnlockDisplay;

    return 0;
}

void _XFreeDisplayLock(Display *dpy)
{
    assert(dpy->lock == NULL);
    if (dpy->lock_fns != NULL) {
	Xfree((char *)dpy->lock_fns);
	dpy->lock_fns = NULL;
    }
}

static void call_handlers(Display *dpy, XCBGenericRep *buf)
{
	_XAsyncHandler *async, *next;
	_XSetLastRequestRead(dpy, (xGenericReply *) buf);
	for(async = dpy->async_handlers; async; async = next)
	{
		next = async->next;
		if(async->handler(dpy, (xReply *) buf, (char *) buf, sizeof(xReply) + (buf->length << 2), async->data))
			return;
	}
	if(buf->response_type == 0) /* unhandled error */
	    _XError(dpy, (xError *) buf);
}

void _XGetXCBBuffer(Display *dpy)
{
    static const xReq dummy_request;

    XCBConnection *c = XCBConnectionOfDisplay(dpy);
    unsigned int *req;

    _xcb_assert_valid_sequence(c);

    _xcb_out_force_sequence_wrap(c);

    dpy->bufptr = dpy->buffer = c->out.queue + c->out.queue_len;
    dpy->bufmax = c->out.queue + sizeof(c->out.queue);

    if(c->out.last_request)
	dpy->last_req = c->out.last_request;
    else
	dpy->last_req = (char *) &dummy_request;

    dpy->request = c->out.request;

    while((req = _xcb_list_peek_head(dpy->xcl->pending_requests)))
    {
	XCBGenericRep *reply;
	XCBGenericError *error;
	/* If this request hasn't been read off the wire yet, save the
	 * rest for later. */
	if((signed int) (c->in.request_read - *req) <= 0)
	    break;
	_xcb_list_remove_head(dpy->xcl->pending_requests);
	/* This can't block due to the above test, but it could "fail"
	 * by returning null for any of several different reasons. We
	 * don't care. In any failure cases, we must not have wanted
	 * an entry in the reply queue for this request after all. */
	reply = XCBWaitForReply(c, *req, &error);
	free(req);
	if(!reply)
	    reply = (XCBGenericRep *) error;
	if(reply)
	    call_handlers(dpy, reply);
	free(reply);
    }

    dpy->last_request_read = c->in.request_read;
    _xcb_assert_sequence_less(dpy->last_request_read, dpy->request);
}

void _XPutXCBBuffer(Display *dpy)
{
    XCBConnection *c = XCBConnectionOfDisplay(dpy);
    int len = dpy->bufptr - dpy->buffer;

    _xcb_assert_sequence_less(dpy->last_request_read, dpy->request);

    if(len)
    {
	struct iovec iov;
	c->out.queue_len += len;
	iov.iov_base = (caddr_t) dpy->buffer;
	iov.iov_len = len;
	_XBeforeFlush(dpy, &iov);
    }

    /* crash applications that use these without locking the display */
    dpy->last_req = dpy->buffer = dpy->bufptr = dpy->bufmax = 0;

    _xcb_assert_sequence_less(c->out.request, dpy->request);
    while(c->out.request != dpy->request)
    {
	unsigned int *req = malloc(sizeof(unsigned int));
	*req = ++c->out.request;
	_xcb_in_expect_reply(c, *req);
	_xcb_list_append(dpy->xcl->pending_requests, req);
    }
    _xcb_assert_valid_sequence(c);
}

/*  */

void _XGetXCBBufferIf(Display *dpy, enum _XBufferCondition locked)
{
    if((dpy->xcl->lock_count > 0) == locked)
	_XGetXCBBuffer(dpy);
}

void _XPutXCBBufferIf(Display *dpy, enum _XBufferCondition locked)
{
    if((dpy->xcl->lock_count > 0) == locked)
	_XPutXCBBuffer(dpy);
}
